module.exports = function(hljs) {
  var IDENT_RE = '[A-Za-z$_][0-9A-Za-z$_]*';
  var KEYWORDS = {
    keyword:
      'in of if for while finally var new function do return void else break catch ' +
      'instanceof with throw case default try this switch continue typeof delete ' +
      'let yield const export super debugger as async await static ' +
      // ECMAScript 6 modules import
      'import from as'
    ,
    literal:
      'true false null undefined NaN Infinity',
    built_in: 'print addLayer centerObject'
      // "ee"+ "Image"+
      // "Algorithms"+ 
      //   "FMask"+ "fillMinima"+ "matchClouds"+
      //   "GeometryConstructors"+ "BBox"+ "LineString"+ "LinearRing"+ "MultiGeometry"+ "MultiLineString"+ "MultiPoint"+ "MultiPolygon"+ "Point"+ "Polygon"+ "Rectangle"+ "Segmentation"+
      //   "Image"+ "Segmentation"+ "GMeans"+ "KMeans"+ "SNIC"+ "seedGrid"+
      //   "Landsat"+ "TOA"+ "calibratedRadiance"+ "pathRowLimit"+ "simpleCloudScore"+ "simpleComposite"+
      //   "Sentinel2"+ "CDI"+
      //   "TemporalSegmentation"+ "Ccdc"+ "Ewmacd"+ "LandTrendr"+ "LandTrendrFit"+ "StructuralChangeBreakpoints"+ "VCT"+ "Verdet"+
      //   "CannyEdgeDetector"+ "Collection"+ "CrossCorrelation"+ "Date"+ "Describe"+ "Dictionary"+ "Feature"+ "HillShadow"+ "HoughTransform"+ "If"+ "IsEqual"+ "ObjectType"+ "Proj"+ "ProjectionTransform"+ "String"+ "Terrain"+
      
      // "Array"+"bitsToArray"+"cat"+"identity"+"abs"+"accum"+"acos"+"add"+"and"+"argmax"+"asin"+"atan"+"atan2"+"bitCount"+"bitwiseAnd"+"bitwiseNot"+"bitwiseOr"+"bitwiseXor"+"byte"+"cbrt"+"ceil"+"cos"+"cosh"+"cut"+"digamma"+"divide"+"dotProduct"+"double"+"eigen"+"eq"+"erf"+"erfInv"+"erfc"+"erfcInv"+"exp"+"first"+"firstNonZero"+"float"+"floor"+"gamma"+"gammainc"+"get"+"gt"+"gte"+"hypot"+"int"+"int16"+"int32"+"int64"+"int8"+"lanczos"+"leftShift"+"length"+"log"+"log10"+"long"+"lt"+"lte"+"mask"+"matrixCholeskyDecomposition"+"matrixDeterminant"+"matrixDiagonal"+"matrixFnorm"+"matrixInverse"+"matrixLUDecomposition"+"matrixMultiply"+"matrixPseudoInverse"+"matrixQRDecomposition"+"matrixSingularValueDecomposition"+"matrixSolve"+"matrixToDiag"+"matrixTrace"+"matrixTranspose"+"max"+"min"+"mod"+"multiply"+"neq"+"not"+"or"+"pad"+"pow"+"project"+"reduce"+"repeat"+"reshape"+"rightShift"+"round"+"short"+"signum"+"sin"+"sinh"+"slice"+"sort"+"sqrt"+"subtract"+"tan"+"tanh"+"toByte"+"toDouble"+"toFloat"+"toInt"+"toInt16"+"toInt32"+"toInt64"+"toInt8"+"toList"+"toLong"+"toShort"+"toUint16"+"toUint32"+"toUint8"+"transpose"+"trigamma"+"uint16"+"uint32"+"uint8"+
      // "Blob"+ "string"+ "url"+ 
      // "Classifier"+ "amnhMaxent"+"decisionTree"+"decisionTreeEnsemble"+"libsvm"+"minimumDistance"+"smileCart"+"smileGradientTreeBoost"+"smileNaiveBayes"+"smileRandomForest"+"spectralRegion"+"confusionMatrix"+"explain"+"mode"+"schema"+"setOutputMode"+"train"+
      // "Clusterer"+//
      // "ConfusionMatrix"+
      // "Date"+
      // "DateRange"+
      // "Dictionary"+
      // "ErrorMargin"+
      // "Feature"+ 
      // "FeatureCollection"+
      // "Filter"+
      // "Geometry"+
      // "Image"+ 
      // "ImageCollection"+ 
      // "Join"+ 
      // "Kernel"+ 
      // "List"+ 
      // "Model"+
      // "Number"+
      // "PixelType"+ 
      // "Projection"+ 
      // "Reducer"+ 
      // "String"+ 
      // "Terrain"+ 
      // "data"+ 
      // "Chart"+
      // "Export"+
      // "Map"+ 
      // "ui"+
      // "Textbox"+ 
      // "exports"+ 
      // "ApiFunction"
  };
  var NUMBER = {
    className: 'number',
    variants: [
      { begin: '\\b(0[bB][01]+)n?' },
      { begin: '\\b(0[oO][0-7]+)n?' },
      { begin: hljs.C_NUMBER_RE + 'n?' }
    ],
    relevance: 0
  };
  var SUBST = {
    className: 'subst',
    begin: '\\$\\{', end: '\\}',
    keywords: KEYWORDS,
    contains: []  // defined later
  };
  var HTML_TEMPLATE = {
    begin: 'html`', end: '',
    starts: {
      end: '`', returnEnd: false,
      contains: [
        hljs.BACKSLASH_ESCAPE,
        SUBST
      ],
      subLanguage: 'xml',
    }
  };
  var CSS_TEMPLATE = {
    begin: 'css`', end: '',
    starts: {
      end: '`', returnEnd: false,
      contains: [
        hljs.BACKSLASH_ESCAPE,
        SUBST
      ],
      subLanguage: 'css',
    }
  };
  var TEMPLATE_STRING = {
    className: 'string',
    begin: '`', end: '`',
    contains: [
      hljs.BACKSLASH_ESCAPE,
      SUBST
    ]
  };
  SUBST.contains = [
    hljs.APOS_STRING_MODE,
    hljs.QUOTE_STRING_MODE,
    HTML_TEMPLATE,
    CSS_TEMPLATE,
    TEMPLATE_STRING,
    NUMBER,
    hljs.REGEXP_MODE
  ];
  var PARAMS_CONTAINS = SUBST.contains.concat([
    hljs.C_BLOCK_COMMENT_MODE,
    hljs.C_LINE_COMMENT_MODE
  ]);

  return {
    aliases: ['js', 'jsx'],
    keywords: KEYWORDS,
    contains: [
      {
        className: 'meta',
        relevance: 10,
        begin: /^\s*['"]use (strict|asm)['"]/
      },
      {
        className: 'meta',
        begin: /^#!/, end: /$/
      },
      hljs.APOS_STRING_MODE,
      hljs.QUOTE_STRING_MODE,
      HTML_TEMPLATE,
      CSS_TEMPLATE,
      TEMPLATE_STRING,
      hljs.C_LINE_COMMENT_MODE,
      hljs.C_BLOCK_COMMENT_MODE,
      NUMBER,
      { // "value" container
        begin: '(' + hljs.RE_STARTERS_RE + '|\\b(case|return|throw)\\b)\\s*',
        keywords: 'return throw case',
        contains: [
          hljs.C_LINE_COMMENT_MODE,
          hljs.C_BLOCK_COMMENT_MODE,
          hljs.REGEXP_MODE,
          {
            className: 'function',
            begin: '(\\(.*?\\)|' + IDENT_RE + ')\\s*=>', returnBegin: true,
            end: '\\s*=>',
            contains: [
              {
                className: 'params',
                variants: [
                  {
                    begin: IDENT_RE
                  },
                  {
                    begin: /\(\s*\)/,
                  },
                  {
                    begin: /\(/, end: /\)/,
                    excludeBegin: true, excludeEnd: true,
                    keywords: KEYWORDS,
                    contains: PARAMS_CONTAINS
                  }
                ]
              }
            ]
          },
          {
            className: '',
            begin: /\s/,
            end: /\s*/,
            skip: true,
          },
          { // E4X / JSX
            begin: /</, end: /(\/[A-Za-z0-9\\._:-]+|[A-Za-z0-9\\._:-]+\/)>/,
            subLanguage: 'xml',
            contains: [
              { begin: /<[A-Za-z0-9\\._:-]+\s*\/>/, skip: true },
              {
                begin: /<[A-Za-z0-9\\._:-]+/, end: /(\/[A-Za-z0-9\\._:-]+|[A-Za-z0-9\\._:-]+\/)>/, skip: true,
                contains: [
                  { begin: /<[A-Za-z0-9\\._:-]+\s*\/>/, skip: true },
                  'self'
                ]
              }
            ]
          }
        ],
        relevance: 0
      },
      {
        className: 'function',
        beginKeywords: 'function', end: /\{/, excludeEnd: true,
        contains: [
          hljs.inherit(hljs.TITLE_MODE, {begin: IDENT_RE}),
          {
            className: 'params',
            begin: /\(/, end: /\)/,
            excludeBegin: true,
            excludeEnd: true,
            contains: PARAMS_CONTAINS
          }
        ],
        illegal: /\[|%/
      },
      {
        begin: /\$[(.]/ // relevance booster for a pattern common to JS libs: `$(something)` and `$.something`
      },
      hljs.METHOD_GUARD,
      { // ES6 class
        className: 'class',
        beginKeywords: 'class', end: /[{;=]/, excludeEnd: true,
        illegal: /[:"\[\]]/,
        contains: [
          {beginKeywords: 'extends'},
          hljs.UNDERSCORE_TITLE_MODE
        ]
      },
      {
        beginKeywords: 'constructor get set', end: /\{/, excludeEnd: true
      }
    ],
    illegal: /#(?!!)/
  };
};